#include <stdio.h>
#include <stdlib.h>
#include <string.h>    /* for memset */
#include <ctype.h>
#include <signal.h>
#include <netdb.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h> /* for inet_ntoa */
#include <unistd.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <termios.h>
#include <fcntl.h>

/* This file is part of Bisqwit's pipes and remotegcc packages. */

static int ConnectOutTimed;
static void SubConnectOut(int signum)
{
	signum = ConnectOutTimed++;
}

static int isinvalidport(const char *s)
{
	char *r;
	strtol(s, &r, 10);	
	return *r ? 1 : 0;
}

static struct termios tattr, saved_attributes;
static void fixmodes(void)
{
    setvbuf(stdout, NULL, _IONBF, 0);

    tcgetattr(STDIN_FILENO, &saved_attributes);
    tcgetattr(STDIN_FILENO, &tattr);
    tattr.c_lflag &= ~ICANON;
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &tattr);
}
static void restoremodes(void)
{
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_attributes);
}
static void cat(void)
{
	char Buf[16384];
	for(;;)
	{
		fd_set fds;
		struct timeval tv = { 60, 0 };
		int i;
		FD_ZERO(&fds);
		FD_SET(0, &fds);
		i = select(1, &fds, NULL, NULL, &tv);
		if(i == 1)
		{
			i = read(0, Buf, sizeof Buf);
			if(i <= 0)break;
			write(1, Buf, i);
		}
	}
	exit(0);
}
#ifdef main
extern int main(int argc, const char *const *argv);
#endif
int main(int argc, const char *const *argv)
{
	int port, fix=0, MyPort=0, verbose=0, usesh=1;
	int Plug, Sock;
	unsigned AddrLen;
	const char *a0 = argv[0], *q;
	const char *connhost = NULL;
	struct sockaddr_in MyAddr, They;
	if(!strcmp(*argv, "cat"))cat();
	while((q = strchr(a0, '/')))a0 = q+1;
	if(argc <= 1)
	{
Usage:
#ifdef main
		printf("Syntax error\n");
#else
		printf(
			"Pipers version "VERSION"\n"
			"Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n\n"
			"Usage: %s [-f] [-v] %s<port> [-ioe <command>] [...]\n"
			"-f = fix console modes first... (termios)\n"
			"-v = verbose (stderr)\n"
			"-s = do not use sh -c, do not even give argv[0]\n"
			"Use capital -switch to fcntl the pipe nonblocking.\n",
			a0, strcmp(a0, "plis") ? "[-p<myport>] <host> " : "");
#endif
		return -1;
	}
	
	for(;;)
	{
		if(!strcmp(argv[1], "-s")) { --argc, ++argv, usesh=0; continue; }
		if(!strcmp(argv[1], "-f")) { --argc, ++argv, fix=1; continue; }
		if(!strcmp(argv[1], "-v")) { --argc, ++argv, ++verbose; continue; }
		break;
	}
	
	if(isinvalidport(argv[1]))
	{
		if(argv[1][0]=='-' && argv[1][1]=='p')
			--argc, MyPort = atoi(*++argv + 2);
		connhost = (--argc, *++argv);
		if(!connhost || argc<=1)goto Usage;
	}
		
	signal(SIGCHLD, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);
	
	port = (--argc, atoi(*++argv));
	
	Plug = socket(AF_INET, SOCK_STREAM, 0);
	if(Plug < 0) { perror("socket"); return errno; }
	
	if(fix)fixmodes();
	
	if(connhost)
	{
		struct hostent *HostName = gethostbyname(connhost);
		if(!HostName) { perror("gethostbyname"); return errno; }
		
		if(MyPort > 0)
		{
		ReBind1:
            MyAddr.sin_port = htons(MyPort);			/* network order */
        	MyAddr.sin_family = AF_INET;				/* host byte order */
        	MyAddr.sin_addr.s_addr = htonl(INADDR_ANY);	/* auto-fill with my IP, network long  */
            memset((void *)&(MyAddr.sin_zero), 0, 8);	/* zero the rest of the struct */
            
            Sock=1; // Just a temp variable, boolean value for setsockopt()
            setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &Sock, sizeof Sock);
            
        	if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(struct sockaddr)))
        	{
        		if(errno == EADDRINUSE)	/* Address already in use */
        		{
        			printf("Port %d busy, sleeping 5 seconds and retrying\n", MyPort);
        			sleep(5);
        			goto ReBind1;
        		}
        		perror("bind");
        		return errno;
        	}
    	}

		memset(&MyAddr, 0, sizeof(MyAddr));
		
		MyAddr.sin_family = AF_INET;
		MyAddr.sin_port = htons(port);
		memcpy(&MyAddr.sin_addr, HostName->h_addr, HostName->h_length);
		signal(SIGABRT, SubConnectOut);
		Sock = fork();
		if(Sock < 0) { perror("fork"); return errno; }
		if(!Sock)
		{
			sleep(5);
			kill(getppid(), SIGABRT);
			_exit(0);
		}
		for(;;)
		{
			if(verbose)
			{
				fprintf(stderr, "Connecting to %s port %d", connhost, port);
				if(MyPort > 0)fprintf(stderr, " (from port %d)", MyPort);
				fprintf(stderr, "...\n");
			}
			if(connect(Plug, (struct sockaddr *)&MyAddr, sizeof(MyAddr)))
			{
				if(errno == EINTR && !ConnectOutTimed)continue;
				if(ConnectOutTimed)errno = ETIMEDOUT;
				perror("connect");
				Plug = -1;
			}
			kill(Sock, SIGKILL);
			signal(SIGABRT, SIG_DFL);
			waitpid(Sock, NULL, WNOHANG); /* Don't harm strace */
			if(Plug < 0)return errno;
			break;
		}
		Sock = Plug;
		goto Diu;
	}

ReBind:
    MyAddr.sin_port = htons(port);				/* network order */
	MyAddr.sin_family = AF_INET;				/* host byte order */
	MyAddr.sin_addr.s_addr = htonl(INADDR_ANY);	/* auto-fill with my IP, network long  */
    memset((void *)&(MyAddr.sin_zero), 0, 8);	/* zero the rest of the struct */
    
    Sock=1; // Just a temp variable, boolean value for setsockopt()
    setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &Sock, sizeof Sock);
    
	if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(struct sockaddr)))
	{
		if(errno == EADDRINUSE)	/* Address already in use */
		{
			printf("Port %d busy, sleeping 5 seconds and retrying\n", port);
			sleep(5);
			goto ReBind;
		}
		perror("bind");
		return errno;
	}
	
	if(listen(Plug, 10))
	{
		perror("listen");
		return errno;
	}
	
	AddrLen = sizeof(They);
	for(;;)
	{
		int i;
		Sock = accept(Plug, (struct sockaddr *)&They, &AddrLen);
		if(Sock < 0)
		{
			perror("accept");
			sleep(1);
			continue;
		} 
		if(verbose)
		{
			fprintf(stderr, "Got connection, from port %d", They.sin_port);
			fprintf(stderr, ", host %s\n", inet_ntoa(They.sin_addr));
		}
		i = fork();
		switch(i)
		{
			case -1:
				perror("fork");
				break;
			case 0:
			{
				int a;
				close(Plug);
Diu:			/* At this point the connection is established. */
				for(a=1; a<argc; a++)
				{
					const char *s = argv[a];
					if(*s=='-')
					{
						int i=0, o=0, e=0, b;
						while(*++s)
							switch(*s)
							{
								case 'i': i=1; break;
								case 'o': o=1; break;
								case 'e': e=1; break;
								case 'I': i=2; break;
								case 'O': o=2; break;
								case 'E': e=2; break;
							}
						++a;
						b = fork();
						if(b < 0)
							perror("fork");
						else if(!b)
						{
							b = Sock;
							if(i) { dup2(b, 0); if(i==2) fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK); }
							if(o) { dup2(b, 1); if(o==2) fcntl(1, F_SETFL, fcntl(1, F_GETFL) | O_NONBLOCK); }
							if(e) { dup2(b, 2); if(e==2) fcntl(2, F_SETFL, fcntl(2, F_GETFL) | O_NONBLOCK); }
							if(!strcmp(argv[a], "cat"))execl(a0, "cat", NULL);
							if(usesh)
								execl(getenv("SHELL"), "sh", "-c", argv[a], NULL);
							else
								execlp(argv[a], NULL);
						}
					}
				}
				if(connhost)
				{
					/* If we're not looping. */
					while(wait(NULL) > 0);
					close(Sock);
					goto Quit;
				}
				_exit(0);
			}
			default:
				close(Sock);
		}
	}
Quit:
	if(fix)restoremodes();
	return 0;
}
